import json
import math
import random
from pathlib import Path
from random import Random

from abrain.core.ann import ANN, Point
from abrain.core.genome import Genome, GIDManager

from _utils import assert_equal


class Robot:
    """A bare-bones individual

    :param genome: the collection of inheritable characteristics
    """

    ann_inputs = None
    ann_outputs = None

    def __init__(self, genome: Genome):
        self.genome = genome
        self.brain = ANN.build(Robot.ann_inputs, Robot.ann_outputs, genome)
        self.fitness = self.brain.stats().depth

    def __repr__(self):
        return f"Robot(id={self.genome.id()}," \
               f" depth={self.fitness}, neurons={len(self.brain.neurons())}," \
               f" axons={self.brain.stats().edges})"

    def to_json(self, file):
        with open(file, 'w') as f:
            json.dump(dict(
                genome=self.genome.to_json(),
                fitness=self.fitness
            ), f)

    @staticmethod
    def from_json_file(file):
        with open(file, 'r') as f:
            data = json.load(f)
            r = Robot(Genome.from_json(data["genome"]))
            r.fitness = data["fitness"]
            return r

    @staticmethod
    def assert_equal(lhs, rhs):
        assert_equal(lhs.genome, rhs.genome)
        assert lhs.fitness == rhs.fitness
        assert len(lhs.brain.neurons()) == len(rhs.brain.neurons())

        lhs_stats, rhs_stats = lhs.brain.stats(), rhs.brain.stats()
        assert lhs_stats.depth == rhs_stats.depth
        assert lhs_stats.edges == rhs_stats.edges
        assert lhs_stats.axons == rhs_stats.axons


def test_evolution(evo_config, capsys, tmp_path):
    pop_size = evo_config["population"]
    gens = evo_config["generations"]
    seed = evo_config["seed"]
    fitness = evo_config["fitness"]

    tour_size = 4
    rng = Random(seed)

    def random_coordinates(n_min, n_max, y):
        def rc(): return random.uniform(-1, 1)
        return [Point(rc(), y, rc()) for _ in
                range(random.randint(n_min, n_max))]

    # By personal convention inputs and outputs are on the y=-1 and y=1 planes,
    # respectively
    Robot.ann_inputs = random_coordinates(10, 20, -1)
    Robot.ann_outputs = random_coordinates(5, 10, 1)

    id_manager = GIDManager()

    with capsys.disabled():
        print()
        print("========================")
        print("==== Starting evolution ====")
        for k, v in evo_config.items():
            print(f"\t{k}: {v}")
        print()

        population = [Robot(Genome.random(rng, id_manager))
                      for _ in range(pop_size)]
        print("== Initialized population")

        def champion(pop=None) -> Robot:
            if pop is None:
                pop = population
            return pop[max(range(len(pop)), key=lambda r: pop[r].fitness)]

        if True:
            first_champ = champion()
            file = f"{tmp_path}/first_champion.json"
            first_champ.to_json(file)

            assert Path(file).exists()

            roundtrip_champ = Robot.from_json_file(file)
            Robot.assert_equal(first_champ, roundtrip_champ)

        for g in range(gens):
            print(f"> Best {fitness}: {champion().fitness}")

            new_population = []
            for _ in range(pop_size):
                winner = champion(rng.sample(population, tour_size))
                child = Robot(winner.genome.mutated(rng, id_manager))
                new_population.append(child)

            population = new_population
            print("== Generated new population [{:0{width}d}] =="
                  .format(g+1, width=math.ceil(math.log10(gens))))

        print("== Final champion ==")
        champ = champion()
        print(f">> {champ}")

        base_path = f"{tmp_path}/champion_gen{gens}"

        json_path = f"{base_path}.json"
        champ.to_json(json_path)
        print(">>> Logged to", json_path)

        roundtrip_champ = Robot.from_json_file(json_path)
        Robot.assert_equal(champ, roundtrip_champ)

        dot_path = champ.genome.to_dot(base_path, "pdf")
        print(">>> Drawn to", dot_path)



import json
import shutil
from pathlib import Path
from typing import Dict

import pytest

from abrain.core.config import Config

default_config = None


@pytest.fixture
def tmp_config_file(tmp_path):
    return tmp_path.joinpath("config.json")


def get_default_config():
    global default_config
    if default_config is None:
        default_config = Config.to_json()
    return default_config


@pytest.fixture
def restore_config_after():
    config = get_default_config()

    yield

    # noinspection PyProtectedMember
    Config.from_json(config)
    Config.show()


def test_config_exists():
    assert Config.functionSet
    assert Config.mutationRates
    assert len(Config._dict) > 0

    for k, v in Config._dict.items():
        print(f"\t{k}: {v}")


@pytest.mark.parametrize("key", ["cppnWeightBounds",
                                 "functionSet", "outputFunctions",
                                 "mutationRates"])
def test_json_conversions(key):
    value = getattr(Config, key)
    fake_json = value.toJson()
    roundtrip = type(value).fromJson(fake_json)

    print(value, fake_json, roundtrip)


def test_config_read_write(tmp_config_file):
    Config.show()
    Config.write(tmp_config_file)
    Config.read(tmp_config_file)
    Config.show()


def test_round_loop(tmp_config_file):
    """
    Execute last in case configuration is screwed up for every one
    :param tmp_config_file:
    :return:
    """
    print(Config.mutationRates)
    print(str(Config.mutationRates))

    original = Config.to_json()

    Config.write(tmp_config_file)

    def header(msg):
        sw, _ = shutil.get_terminal_size((80, 20))
        print(f"{'_'*sw}\n== {msg} ==\n")

    def show(msg):
        header(msg)
        Config.show()
        print()

    show("Default file contents")

    header("On-disk config file")
    print(tmp_config_file.read_text())
    print()

    header("Resetting values")
    for section in Config._sections.values():
        for item in section:
            attr = getattr(Config, item)
            default = type(attr)()
            print(f"\t{item}: {attr} of type {type(attr)} >> {default}")
            setattr(Config, item, default)
    print()

    show("Reset values")

    Config.read(tmp_config_file)

    show("Restored values")

    assert Config.to_json() == original


def test_failed_write(tmp_config_file):
    Config.write(tmp_config_file)
    with pytest.raises(IOError, match="exist"):
        Config.write(tmp_config_file)


def test_failed_read(tmp_config_file):
    with pytest.raises(IOError, match="not exist"):
        Config.read(Path(f"{tmp_config_file}/not-existing-sub-folder"))


def nested_getter(key: str, j: Dict):  # pragma: no cover
    for section, items in Config._sections.items():
        for k in items:
            if k == key:
                return j[section][key]
    raise KeyError


def nested_setter(key: str, value, j: Dict):  # pragma: no cover
    for section, items in Config._sections.items():
        for k in items:
            if k == key:
                j[section][key] = value
                return
    raise KeyError


def change_config_value(key, value, path):
    contents = Config.to_json()

    # print("--"*80)
    # print(f"Changing {key}: {value}")
    # print(contents)

    nested_setter(key, value, contents)

    # print(contents)

    with open(path, 'w') as outfile:
        json.dump(contents, outfile)
        # print("Writing altered configuration to", outfile.name)
    Config.read(path)
    # Config.show()


@pytest.mark.usefixtures("restore_config_after")
@pytest.mark.parametrize(
    'key, value',
    [
        pytest.param("iterations", 15, id="int->int"),
        pytest.param("annWeightsRange", 1.5, id="float->float"),
        pytest.param("activationFunc", "id", id="str->str"),
        pytest.param("cppnWeightBounds", [1, 1, 1, 1, 1],
                     id="List->Bounds"),
        pytest.param("functionSet", ["id", "abs", "ssgn"], id="List->Strings"),
        pytest.param("mutationRates",
                     nested_getter("mutationRates", get_default_config()),
                     id="MutationRates->MutationRates")
    ])
def test_correct_read_type(key, value, tmp_config_file):
    change_config_value(key, value, tmp_config_file)


@pytest.mark.usefixtures("restore_config_after")
@pytest.mark.parametrize(
    'key, value',
    [
        pytest.param("annWeightsRange", "error", id="str->float"),
        pytest.param("iterations", 4.2, id="float->int"),
        pytest.param("cppnWeightBounds", "foo", id="str->Bounds"),
        pytest.param("functionSet", "42", id="str->Strings"),
        pytest.param("mutationRates", ["foo", "bar"], id="bad_mr"),
    ])
def test_failed_read_type(key, value, tmp_config_file):
    with pytest.raises(TypeError):
        change_config_value(key, value, tmp_config_file)


@pytest.mark.usefixtures("restore_config_after")
@pytest.mark.parametrize(
    'key, value',
    [
        pytest.param("activationFunc", "sin", id="act"),
        pytest.param("functionSet", ["abs", "id"], id="funcs"),
        pytest.param("outputFunctions", ["id", "id", "id"], id="outputs")
    ])
def test_correct_read_depends(key, value, tmp_config_file):
    change_config_value(key, value, tmp_config_file)


@pytest.mark.usefixtures("restore_config_after")
@pytest.mark.parametrize(
    'key, value',
    [
        pytest.param("activationFunc", "1", id="act"),
        pytest.param("functionSet", ["circle_quadrature"], id="funcs"),
        pytest.param("outputFunctions", ["id", "id"], id="outputs"),
        pytest.param("outputFunctions", ["sa", "la", "mi"], id="outputs"),
        pytest.param("cppnWeightBounds", [3, 1, -1, -3, .1],
                     id="inv_bounds"),
        pytest.param("cppnWeightBounds", [-3, -1, 1, 3, -.1],
                     id="neg_bounds"),
        pytest.param("mutationRates", {}, id="empty_mr"),
        pytest.param("mutationRates", {"add_n": 0}, id="null_mr"),
        pytest.param("mutationRates", {"add_n": -1}, id="neg_mr"),
    ])
def test_failed_read_depends(key, value, tmp_config_file):
    with pytest.raises(ValueError):
        change_config_value(key, value, tmp_config_file)


def test_activation_function():
    assert Config.activationFunc == 'ssgn'
    print(Config.activationFunc)



from typing import Optional, Dict

import plotly.graph_objects as go

from .._cpp.phenotype import ANN, Point


def plotly_render(ann: ANN, labels: Optional[Dict[Point, str]] = None) \
        -> go.Figure:
    """
    Produce a 3D figure from an artificial neural network

    The returned figure can be used to save an interactive html session or a
    (probably poorly) render to e.g. a png file
    """

    n_type = ANN.Neuron.Type
    nodes = {t: [[], [], []] for t in [n_type.I, n_type.H, n_type.O]}
    edges = [[], [], []]

    names = {t: [] for t in nodes}

    for n in ann.neurons():
        x, y, z = n.pos.tuple()

        n_list = nodes[n.type]
        n_list[0] += [x]
        n_list[1] += [y]
        n_list[2] += [z]

        if labels is None or (label := labels.get(n.pos, None)) is None:
            label = f"{n.type.name}{len(names[n.type])}"
        names[n.type].append(label)

        for link in n.links():
            x2, y2, z2 = link.src().pos.tuple()
            edges[0] += [x, x2, None]
            edges[1] += [y, y2, None]
            edges[2] += [z, z2, None]

    trace_edges = go.Scatter3d(
        x=edges[0], y=edges[1], z=edges[2],
        mode='lines', line=dict(color='black', width=1), opacity=.1,
        hoverinfo='none', showlegend=False
    )

    trace_nodes = []
    for n_type, name, color in [(n_type.I, 'Inputs',  'black'),
                                (n_type.O, 'Outputs', 'black'),
                                (n_type.H, 'Hidden',  'gray')]:
        n_list = nodes[n_type]
        trace_nodes.append(
            go.Scatter3d(
                x=n_list[0], y=n_list[1], z=n_list[2],
                mode='markers',
                marker=dict(symbol='circle',
                            size=10,
                            color=color),
                name=name,
                hovertemplate='<b>' + name + '</b><br>'
                              '<i>%{hovertext}</i><br>'
                              'x: %{x}<br>'
                              'y: %{y}<br>'
                              'z: %{z}<extra></extra>',
                hovertext=names[n_type]
            )
        )

    data = [trace_edges, *trace_nodes]
    fig = go.Figure(data=data)

    camera = dict(
        eye=dict(x=-2, y=0, z=1)
    )

    fig.update_layout(
        scene=dict(
            xaxis=dict(nticks=3, range=[-1.1, 1.1]),
            yaxis=dict(nticks=3, range=[-1.1, 1.1]),
            zaxis=dict(nticks=3, range=[-1.1, 1.1])
        ), scene_aspectmode='cube',
        scene_camera=camera,
        showlegend=False,
        hoverlabel=dict(
            bgcolor="white",
            font_size=16,
            font_family="Courrier"
        )
    )

    return fig



import json
import logging
import pprint
from pathlib import Path
from typing import Optional, Dict

from .._cpp.config import Config as CPPConfig
from .._cpp.phenotype import CPPN

logger = logging.getLogger(__name__)


class Config(CPPConfig):
    """
    Wrapper for the C++ configuration data

    Allows transparent access to the values used for CPPN structure,
    genome mutation (:func:`Genome.mutate`), and ANN/ES-HyperNEAT parameters
    """

    #: The set of all wrapped C++ values
    _dict = {
        k: v for k, v in CPPConfig.__dict__.items() if not k.startswith('_')
    }

    #: Private reference to the config sections
    _sections = CPPConfig.__dict__['_sections']

    @classmethod
    def to_json(cls) -> Dict:
        """
        Convert to a json-compliant Python dictionary
        """
        dct = {}

        for section, items in cls._sections.items():
            dct[section] = {}
            for k in items:
                attr = getattr(cls, k)
                if hasattr(attr, "toJson"):
                    attr = attr.toJson()
                dct[section][k] = attr

        return dct

    @classmethod
    def from_json(cls, j: Dict):
        """
        Restore values from a json-compliant Python dictionary

        :param j: the dictionary to parse values from
        """
        for section, items in cls._sections.items():
            for k in items:
                attr = j[section][k]
                this_attr = getattr(cls, k)
                if hasattr(this_attr, "fromJson"):
                    j_attr = attr
                    attr = type(this_attr).fromJson(j_attr)

                    if not attr.isValid():
                        raise ValueError(
                            f"{attr} is not a valid value for {k}")
                try:
                    setattr(cls, k, attr)
                except TypeError as e:
                    raise TypeError(f"Failed to convert {attr}:"
                                    f" {type(attr)} -> {type(this_attr)}\n{e}")

    @classmethod
    def write(cls, path: Optional[Path]):
        """
        Write the configuration to the specified file or stdout

        :param path: where to write or none to print to screen
        """

        if path is not None and path.exists():
            raise IOError(f"Will not overwrite existing "
                          f"file {path} with this one")

        json_cls = cls.to_json()

        if path is None:
            pprint.pprint(json_cls)
        else:
            with open(path, 'w') as f:
                json.dump(json_cls, f)
            assert path.exists()

    @classmethod
    def show(cls):
        """
        Write the configuration on standard output
        """

        cls.write(path=None)

    @classmethod
    def read(cls, path: Path):
        """
        Try to load data from provided path

        :param path: Filename
        """

        if not path.exists():
            raise IOError(f"Input path '{path}' does not exist")

        with open(path, 'r') as f:
            cls.from_json(json.load(f))

        # Then check relationship assertions
        if not cls.known_function(cls.activationFunc):
            raise ValueError(f"Activation function {cls.activationFunc}"
                             f" is unknown")

        for func in cls.functionSet:
            if not cls.known_function(func):
                raise ValueError(f"CPPN internal function {func} is unknown")

        for func in cls.outputFunctions:
            if not cls.known_function(func):
                raise ValueError(f"CPPN output function {func} is unknown")

        if len(cls.outputFunctions) != CPPN.OUTPUTS:
            raise ValueError(f"Invalid number of output functions in"
                             f" {cls.outputFunctions}. CPPN requires"
                             f" {CPPN.OUTPUTS}")

        # Let us assume I am careful enough for the other values

        return path



import copy
import logging
import pickle
import pydoc
from pathlib import Path
from random import Random
from typing import Optional

import pytest
from abrain.core.config import Config
from abrain.core.genome import Genome, logger as genome_logger, GIDManager

from _utils import assert_equal

logging.root.setLevel(logging.NOTSET)
logging.getLogger('graphviz').setLevel(logging.WARNING)

###############################################################################
# Generic tests
###############################################################################


def test_exists():
    rng = Random(16)
    g = Genome.random(rng)
    print(g)
    print(pydoc.render_doc(g))


def test_default_genome_fails():
    with pytest.raises(AssertionError):
        Genome()


def test_random_genome():
    g = Genome.random(Random(16))
    print(g)
    assert "CPPN" in g.__repr__(), \
        f"Wrong output format for CPPN genome description. Got {g.__repr__()}"

    size_before = len(g.links)
    g._add_link(0, 0, 0)
    assert len(g.links) == size_before + 1, "Failed to append a link"

    g._add_node("id")
    assert len(g.nodes) == 1

    print(f"Full details of {g}")
    print("\tNodes:")
    for node in g.nodes:
        print(f"\t\t{node}")
    print("\tLinks:")
    for link in g.links:
        print(f"\t\t{link}")


###############################################################################
# Mutation tests
###############################################################################

class RatesGuard:
    def __init__(self, rates: dict):
        self.override = rates

    def __enter__(self):
        # print("Backed mutation rates:", Config.mutationRates)
        self.backup = {k: v for k, v in Config.mutationRates.items()}
        for k in Config.mutationRates.keys():
            Config.mutationRates[k] = 0
        for k, v in self.override.items():
            Config.mutationRates[k] = v
        # print("Replaced mutation rates with:", Config.mutationRates)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for k, v in self.backup.items():
            Config.mutationRates[k] = v
        # print("Restored mutation rates:", Config.mutationRates)


def test_mutate_genome_create(tmp_path):
    g = Genome.random(Random(16))
    assert g.nextNodeID == 0


def save_function(g: Genome, path: Path, capfd):
    def helper(gen: Optional[int] = None, title: Optional[str] = None):
        if gen is not None:
            setattr(helper, 'gen', gen)

        g.to_dot(path=f"{path}/gen{getattr(helper, 'gen'):02}",
                 ext="png", debug="depths", title=title)

        captured = capfd.readouterr()
        assert len(captured.err) == 0

        setattr(helper, 'gen', getattr(helper, 'gen') + 1)
    return helper


def test_mutate_genome_add(tmp_path, capfd):
    rng = Random(16)
    g = Genome.random(rng)
    save = save_function(g, tmp_path, capfd)
    with RatesGuard({"add_n": 1}):
        save(0)
        steps = 10
        for i in range(steps):
            g.mutate(rng)
            save()
        assert len(g.nodes) == steps
        assert g.nextNodeID == steps

    with RatesGuard({"add_l": 1}):
        nl = len(g.links)
        save()
        steps = 10
        for i in range(steps):
            g.mutate(rng)
            save()
        assert len(g.links) == nl + steps


def test_mutate_genome_del_n(tmp_path, capfd):
    steps = 10
    rng = Random(16)
    g = Genome.random(rng)
    save = save_function(g, tmp_path, capfd)
    with RatesGuard({"add_n": 1}):
        save(0)
        for i in range(steps):
            g.mutate(rng)
            save()
    with RatesGuard({"del_n": 1}):
        save()
        for i in range(steps):
            g.mutate(rng)
            save()
        assert len(g.nodes) == 0
        assert g.nextNodeID == steps


def test_mutate_genome_del_l(tmp_path, capfd):
    steps = 10
    rng = Random(16)
    g = Genome.random(rng)
    save = save_function(g, tmp_path, capfd)
    save(0)

    with RatesGuard({"add_n": 1}):
        for i in range(steps):
            g.mutate(rng)
            save()

    init_len = len(g.links)
    init_lid = g.nextLinkID
    with RatesGuard({"add_l": 1}):
        for i in range(steps):
            g.mutate(rng)
            save()
    assert len(g.links) == init_len+steps
    assert g.nextLinkID == init_lid+steps

    with RatesGuard({"del_l": 1}):
        for i in range(steps):
            g.mutate(rng)
            save()
    assert len(g.links) == init_len
    assert g.nextLinkID == init_lid+steps


def test_mutate_genome_mut(tmp_path, capfd):
    rng = Random(16)
    g = Genome.random(rng)
    save = save_function(g, tmp_path, capfd)

    save(0)
    with RatesGuard({"add_n": 1}):
        g.mutate(rng)
        save()

    assert len(g.nodes) == 1

    with RatesGuard({"mut_f": 1}):
        steps = 10
        for i in range(steps):
            func = str(next(iter(g.nodes)).func)
            g.mutate(rng)
            save()
            assert str(next(iter(g.nodes)).func) != func

    with RatesGuard({"mut_w": 1}):
        steps = 10
        for i in range(steps):
            weights = [l_.weight for l_ in g.links]
            g.mutate(rng)
            save()
            assert any(b_w != a_w for b_w, a_w in
                       zip(weights, [l_.weight for l_ in g.links]))


def mutate_genome_topology(ad_rate, seed, tmp_path, output, gens, capfd) \
        -> Genome:
    rates = {k: 0 if k.startswith("mut") else 1 for k in Config.mutationRates}
    rates["add_l"] = ad_rate
    rates["add_n"] = ad_rate
    rates["del_l"] = 1/ad_rate
    rates["del_n"] = 1/ad_rate

    title = f"add/del: {ad_rate}, seed: {seed}"

    with RatesGuard(rates):
        rng = Random(seed)
        g = Genome.random(rng)

        if output:
            save = save_function(g, tmp_path, capfd)
            save(0, title)

        for j in range(gens):
            try:
                g.mutate(rng)
                assert len(g.nodes) == len(set([n.id for n in g.nodes]))
                assert len(g.links) == len(set([(li.src, li.dst) for li
                                                in g.links]))
                assert not any(li.src == li.dst for li in g.links)

                def valid_nid(nid): return nid < \
                    g.nextNodeID + g.INPUTS + g.OUTPUTS
                # noinspection PyProtectedMember
                degrees = Genome._compute_node_degrees(g.links)
                for node in g.nodes:
                    d = degrees[node.id]
                    assert not Genome._is_hidden(node.id) \
                           or (d.i > 0 and d.o > 0)
                    assert valid_nid(node.id)
                for link in g.links:
                    assert link.id < g.nextLinkID
                    assert valid_nid(link.src)
                    assert valid_nid(link.dst)

            except Exception as e:  # pragma: no cover
                genome_logger.debug("== Exception ==")
                genome_logger.debug(f">> {e}")
                genome_logger.debug("== Post fail details ==")
                genome_logger.debug("Nodes:")
                # noinspection PyProtectedMember
                degrees = Genome._compute_node_degrees(g.links)
                # noinspection PyProtectedMember
                depths = Genome._compute_node_depths(g.links)
                for node in sorted(g.nodes, key=lambda n_: n_.id):
                    d = degrees[node.id]
                    genome_logger.debug(f"\t{node} i:{d.i}, o:{d.o},"
                                        f" d:{depths[node.id]}")
                genome_logger.debug("Links:")
                for link in sorted(g.links, key=lambda l_: l_.id):
                    genome_logger.debug(f"\t{link}")

                g.to_dot(f"{tmp_path}/faulty_graph", "png", debug=True)
                genome_logger.info(f"Wrote faulty graph to {tmp_path}/"
                                   f"faulty_graph.png")
                genome_logger.info(f"Wrote log to {tmp_path}/log")

                raise e

            if output:
                save(title=title)

        return g


def test_mutate_genome_topology_with_gvc_output(ad_rate, seed, tmp_path,
                                                capfd):
    mutate_genome_topology(ad_rate, seed, tmp_path, output=True, gens=100,
                           capfd=capfd)


def test_mutate_genome_topology(ad_rate, seed, tmp_path, capfd):
    gens = 1000
    g = mutate_genome_topology(ad_rate, seed, tmp_path, output=False,
                               gens=gens, capfd=None)
    save_function(g, tmp_path, capfd)(gens-1,
                                      title=f"gen{gens-1}, add/del: {ad_rate},"
                                            f" seed: {seed}")


def test_mutate_genome_deepcopy(seed):
    rng = Random(seed)
    parent = Genome.random(rng)

    steps = 10

    with RatesGuard({"add_n": 1}):
        for _ in range(steps):
            parent.mutate(rng)
        child = parent.mutated(rng)

        assert_equal(parent, parent.copy())

    assert len(parent.nodes) == len(child.nodes) - 1

    with RatesGuard({"add_n": 1}):
        for _ in range(steps):
            parent.mutate(rng)

    with RatesGuard({"del_n": 1}):
        for _ in range(steps+1):
            child.mutate(rng)
        assert len(child.nodes) == 0

    assert len(parent.nodes) == 2*steps
    assert len(child.nodes) == 0


###############################################################################
# Serialization tests
###############################################################################

def _simple_genome(seed, with_id):
    rng = Random(seed)
    id_manager = GIDManager() if with_id else None
    genome = Genome.random(rng, id_manager)
    for _ in range(10):
        genome.mutate(rng)
    return genome


@pytest.mark.parametrize('with_id', [True, False])
def test_pickle_genome(seed, with_id):
    genome = _simple_genome(seed, with_id)
    roundabout = pickle.loads(pickle.dumps(genome))
    assert_equal(genome, roundabout)


@pytest.mark.parametrize('with_id', [True, False])
def test_json_genome(seed, with_id):
    genome = _simple_genome(seed, with_id)
    roundabout = Genome.from_json(genome.to_json())
    assert_equal(genome, roundabout)


@pytest.mark.parametrize('with_id', [True, False])
def test_copy_genome(seed, with_id):
    genome = _simple_genome(seed, with_id)
    copied = genome.copy()
    assert_equal(genome, copied)


@pytest.mark.parametrize('with_id', [True, False])
def test___copy___genome(seed, with_id):
    genome = _simple_genome(seed, with_id)
    copied = copy.copy(genome)
    assert_equal(genome, copied)


@pytest.mark.parametrize('with_id', [True, False])
def test___deepcopy___genome(seed, with_id):
    genome = _simple_genome(seed, with_id)
    copied = copy.deepcopy(genome)
    assert_equal(genome, copied)



import logging
import math
from random import Random
from time import perf_counter
from typing import Tuple, List, Dict

import pytest

from abrain import Point, Config, Genome, ANN, plotly_render
from abrain.core.genome import logger as genome_logger


def test_default_is_empty():
    ann = ANN()
    assert ann.empty()
    assert len(ann.ibuffer()) == 0
    assert len(ann.obuffer()) == 0

    stats = ann.stats()
    assert stats.depth == 0
    assert stats.edges == 0
    assert stats.axons == 0

    with pytest.raises(ValueError):
        p0 = Point(0, 0, 0)
        print("Testing no point at", p0)
        ann.neuronAt(p0)


def _make_ann(mutations, rng, other_inputs=None, other_outputs=None) -> \
        Tuple[ANN, List[Point], List[Point]]:
    genome = Genome.random(rng)
    for _ in range(mutations):
        genome.mutate(rng)

    def d(): return rng.randint(10, 20)
    def c(): return rng.uniform(-1, 1)
    def p(y): return Point(c(), y, c())

    def append(lhs, rhs):
        if rhs is not None:
            lhs += [Point(x, y, z) for x, y, z in rhs]

    inputs = [p(-1) for _ in range(d())]
    append(inputs, other_inputs)

    outputs = [p(1) for _ in range(d())]
    append(outputs, other_outputs)

    return ANN.build(inputs, outputs, genome), inputs, outputs


def test_empty_perceptrons(mutations, seed):
    n = 10

    def generate_stats():
        rng = Random(seed)
        l_stats: Dict = {key: 0 for key in ['empty', 'perceptron', 'ann']}
        for _ in range(n):
            ann, _, _ = _make_ann(mutations, rng)
            l_stats['empty'] += ann.empty()
            l_stats['perceptron'] += ann.perceptron()
            l_stats['ann'] += (not ann.empty(strict=True))
            assert not ann.empty() or ann.empty(strict=True)
            assert not ann.perceptron() or ann.empty(strict=True)
            assert not (ann.empty() and ann.perceptron())
            assert not ann.perceptron() or Config.allowPerceptrons

        return l_stats

    genome_logger.setLevel(logging.CRITICAL)
    allow_perceptrons = bool(Config.allowPerceptrons)
    Config.allowPerceptrons = True
    stats_t = generate_stats()

    Config.allowPerceptrons = False
    stats_f = generate_stats()
    Config.allowPerceptrons = allow_perceptrons

    print(f"{'-':11s} {'True':10s} {'False':10s}")
    for k in stats_t:
        print(f"{k:10s}", end='')
        for stats in [stats_t, stats_f]:
            print(f" {100*stats[k]/n:10g}%", end='')
        print()

    for s in [stats_t, stats_f]:
        assert (sum(s.values()) == n)
    assert stats_t['empty'] <= stats_f['empty']
    assert stats_f['perceptron'] == 0


@pytest.mark.parametrize(
    'inputs, outputs',
    [
        pytest.param([(0, 0, 0), (0, 0, 0)], [(1, 1, 1)], id="in"),
        pytest.param([(1, 1, 1)], [(0, 0, 0), (0, 0, 0)], id="out"),
        pytest.param([(0, 0, 0)], [(0, 0, 0)], id="io"),
    ])
def test_invalid_duplicates(inputs, outputs):
    with pytest.raises(ValueError):
        print(f"ANN.build({inputs}, {outputs}, genome)")
        _make_ann(0, Random(0), inputs, outputs)


# .. todo:: implement (code exists in ann.cpp)
# def test_deepcopy():
#     original = _make_ann(1000, 0)
#     original.cop


def test_random_eval(mutations, seed):
    rng = Random(seed)
    ann, inputs, outputs = _make_ann(mutations, rng)

    assert len(ann.ibuffer()) == len(inputs)
    assert len(ann.obuffer()) == len(outputs)
    assert ann.empty() or ann.stats().edges > 0

    assert ann.ibuffer() == ann.ibuffer()  # same objects
    assert ann.obuffer() == ann.obuffer()  #

    avg_output = 0
    for _ in range(1000):
        # inputs, outputs = ann.buffers()
        inputs, outputs = ann.ibuffer(), ann.obuffer()
        for j in range(len(inputs)):
            inputs[j] = rng.uniform(-1, 1)
        if hasattr(outputs, "set_to_nan"):  # pragma: no cover
            outputs.set_to_nan()

        ann(inputs, outputs, 1)

        for i in range(len(outputs)):
            avg_output += outputs[i]

        if hasattr(outputs, "set_to_nan"):  # pragma: no cover
            assert ann.empty() or not \
                any([math.isnan(outputs[i]) for i in range(len(outputs))])

    assert ann.empty() or sum != 0


def test_view_neurons_png(mutations, seed, tmp_path):
    rng = Random(seed)
    ann, _, _ = _make_ann(mutations, rng)

    file = f"{tmp_path}/ann.png"
    fig = plotly_render(ann)

    try:
        fig.write_image(file)
        print("Generated", file)    # pragma: no cover
    except Exception as e:  # pragma: no cover
        pytest.skip(f"Ignoring exceptions from unstable kaleido: "
                    f"{e=}, {type(e)=}")


@pytest.mark.parametrize('mutations', [10])
@pytest.mark.parametrize('seed', [1])
@pytest.mark.parametrize('with_labels', [True, False])
def test_view_neurons_interactive(mutations, seed, with_labels, tmp_path):
    start = perf_counter()

    def time():
        nonlocal start
        duration = perf_counter() - start
        start = perf_counter()
        return duration

    rng = Random(seed)
    ann, inputs, _ = _make_ann(mutations, rng)
    print(f"Generating ANN(gen={mutations}, seed={seed}): {time()}s")

    labels = None
    if with_labels:
        labels = {}
        for i in inputs:
            labels[i] = f"Input{len(labels)}"

    fig = plotly_render(ann, labels)
    print(f"Preparing rendering: {time()}s")

    fig.write_html(f"{tmp_path}/interactive.ann.html")
    print(f"Writing: {time()}s")



import array
import pydoc
from itertools import chain, combinations
from random import Random

from abrain._cpp.phenotype import (  # noqa
    CPPN, Point,
)
from abrain.core.genome import Genome
from pytest_steps import test_steps


def _make_cppn(seed, mutations=0):
    rng = Random(seed)
    genome = Genome.random(rng)
    for _ in range(mutations):
        genome.mutate(rng)
    try:
        cppn = CPPN(genome)
    except Exception as e:  # pragma: no cover
        genome.to_dot("foo.pdf", debug="depths,links")
        raise e
    return cppn, genome


def test_exists():
    cppn = _make_cppn(16)
    print(cppn)
    print(pydoc.render_doc(cppn))
    print(pydoc.render_doc(CPPN.Output))
    print(CPPN.Output.Weight)
    print(CPPN.OUTPUTS_LIST)
    assert CPPN.Output.Weight in CPPN.OUTPUTS_LIST


def test_functions():
    for k, v in CPPN.functions().items():
        print(f"{k}(0) = {v(0)}")


def test_create(seed):
    cppn = _make_cppn(seed)
    print(cppn)


def test_output_single(seed):
    cppn, _ = _make_cppn(seed)
    p0, p1 = Point(0., 0., 0.), Point(0., 0., 0.)
    for o in CPPN.OUTPUTS_LIST:
        values = set()
        for i in range(100):
            values.add(cppn(p0, p1, o))
        assert len(values) == 1


def test_outputs_all(seed):
    cppn, _ = _make_cppn(seed)
    p0, p1 = Point(0., 0., 0.), Point(0., 0., 0.)
    values = {k: set() for k in CPPN.OUTPUTS_LIST}
    outputs = CPPN.outputs()
    for i in range(100):
        cppn(p0, p1, outputs)
        for o in CPPN.OUTPUTS_LIST:
            values[o].add(outputs[o])

    for o in CPPN.OUTPUTS_LIST:
        assert len(values[o]) == 1


def test_outputs_subset(seed):
    cppn, _ = _make_cppn(seed)
    p0, p1 = Point(0., 0., 0.), Point(0., 0., 0.)
    values = {k: set() for k in CPPN.OUTPUTS_LIST}
    subset = {CPPN.Output.Weight, CPPN.Output.LEO}
    outputs = CPPN.outputs()
    for i in range(100):
        cppn(p0, p1, outputs, subset)
        for o in CPPN.OUTPUTS_LIST:
            values[o].add(outputs[o])

    for o in CPPN.OUTPUTS_LIST:
        assert len(values[o]) == 1


def test_outputs_equals(seed):
    cppn, _ = _make_cppn(seed)

    rng = Random(seed)

    def r():
        return rng.uniform(-1, 1)

    def p():
        return Point(r(), r(), r())

    def all_subsets(ss):
        return list(chain(*map(lambda x: combinations(ss, x),
                               range(0, len(ss)+1))))

    for _ in range(100):
        p0, p1 = p(), p()

        outputs_manual = []
        for o in CPPN.OUTPUTS_LIST:
            outputs_manual.append(cppn(p0, p1, o))

        outputs = CPPN.outputs()
        # Test all combinations
        outputs_subsets = []
        for subset in all_subsets(CPPN.OUTPUTS_LIST):
            cppn(p0, p1, outputs, set(subset))
            outputs_subsets.append((subset,
                                    [outputs[i] for i in range(len(outputs))]))

        cppn(p0, p1, outputs)
        outputs_all = [outputs[i] for i in range(len(outputs))]

        assert outputs_manual == outputs_all
        for subset, outputs_subset in outputs_subsets:
            for i in range(len(outputs_manual)):
                if CPPN.Output(i) in subset:
                    assert outputs_manual[i] == outputs_subset[i]


sample_sizes = [10, 50, 100]


@test_steps('generate_cppn', *[f"sample_at_{s}" for s in sample_sizes])
def test_multiscale_sampling(mutations, seed,
                             tmp_path_factory):  # pragma: no cover
    folder = tmp_path_factory.mktemp(numbered=False,
                                     basename=f"test_multiscale_sampling_"
                                              f"s{seed}_m{mutations}")

    cppn, genome = _make_cppn(seed, mutations)

    genome.to_dot(f"{folder}/cppn.png")
    print('Generated', f"{folder}/cppn.png")

    yield 'generate_cppn'

    dp = Point(0, 0, 0)
    for size in sample_sizes:
        def to_substrate_coord(index):
            return 2 * index / (size - 1) - 1

        data = array.array('B', [0, 0, 0] * size * size * 2)

        outputs = CPPN.outputs()

        for i in range(size):
            x = to_substrate_coord(i)
            for j in range(size):
                y = to_substrate_coord(j)
                p = Point(x, y, 0)
                for k, p0, p1 in [(0, p, dp), (1, dp, p)]:
                    cppn(p, dp, outputs)
                    ix = 3*(i+size*(j+size*k))
                    for o, v in enumerate(
                            [outputs[i] for i in range(len(outputs))]):
                        data[ix+o] = \
                            int(.5 * (max(-1.0, min(outputs[o], 1.0)) + 1)
                                * 255)

        for name, i in [('input', 0), ('output', 1)]:
            filename = f"{folder}/xy_{name}_plane_{size:03}x{size:03}.ppm"
            with open(filename, 'wb') as f:
                f.write(bytearray(f"P6\n{size} {size}\n255\n", 'ascii'))
                data.tofile(f)
            print('Generated', filename)

        yield f"sample_at_{size}"
